# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: fly.sun <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/01/19
#
"""
我们自己实现一个 layerNorm， 注意 layerNorm 是对每一条数据进行 Norm，而不是每一批数据，这两个很像，但是作用纬度不一样。
在 NLP 任务中，我们使用 layerNorm 比较多，因为是：

1、文本自身是变长的，max_length 为 512 的话，可能大部分的数据都只有几十个字。那么让这几十个字以及大批的 padding 进行 norm 是不合理的。
2、batchNorm 中的 平均值 和 方差，是在训练任务中学到的。 然后推理的时候，根据训练任务中学到的平均值和方法来使用，比如 cv 中常见的 transforms.Normalize。
如果使用 layerNorm 的话，就不需要提前计算好平均值和方法，每句话输入进来的时候，单独计算就可以了。
对于变长文本预测来说，这样其实更合理一些。
3、自己实现 layerNorm 还可以方便后续进行一些细小的优化。
4、参考 https://iii.run/archives/fae41911210f.html 实现

"""

import torch
import torch.nn as nn


class LayerNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-12):
        super(LayerNorm, self).__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.eps = eps

    def forward(self, x):
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight * x + self.bias



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: fly.sun <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/01/24
#
""""""

import logging
import os
import sys

_log_format = "%(asctime)s |- %(levelname).1s %(name)s - %(message)s"


def _get_log_level():
    """ get log level from env variable 'LOG_LEVEL'

    Returns:
        str|int: e.g. "INFO", 20, "DEBUG", 10, "ERROR", 40.

    """
    level = os.environ.get("LOG_LEVEL", "INFO")
    try:
        level = int(level)
    except ValueError:
        assert isinstance(level, str)
        level = level.upper()
    return level


# Attention do not change
# unless you test it online and there is no DEBUG log
# add stream handler to avoid default stream heandler NOSET level
logging.basicConfig(
    format=_log_format,
    level=_get_log_level(),
    stream=sys.stdout)


class LogLevel(object):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    DETAIL = 5
    NOTSET = 0


def init_logger(logger_name=None,
                log_file=os.environ.get("LOG_FILE", ""),
                log_format=_log_format,
                level=_get_log_level()):
    """ init logger

    Args:
        logger_name(str): optional, default: None.
        log_file(str): optional, default: "".
            output log messages to file if specified, by default is set by env
            `LOG_FILE`.
        log_format(str): optional, default:
            "%(asctime)s |- %(levelname).1s %(name)s - %(message)s"
        level(int|logging.Level): set log level, by default it is set by env
            `LOG_LEVEL`, `INFO` level is used if not set.
        :: level
            - CRITICAL    50
            - ERROR	40
            - WARNING	30
            - INFO	20
            - DEBUG	10
            - DETAIL  5
            - NOTSET  0

    Returns:
        logging.Logger: a logger instance

    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    if log_file:
        handler = logging.FileHandler(log_file)
        if log_format:
            formatter = logging.Formatter(log_format)
            handler.setFormatter(formatter)

        logger.addHandler(handler)

    return logger


def _test():
    logger = init_logger("test_logger", "test_file.log",
                         level=_get_log_level())
    logger.info("level: {}".format(os.environ.get("LOG_LEVEL", "INFO")))
    import sys
    logger.info(sys.modules[__name__])
    logger.info(logging.getLoggerClass())
    logger.debug("test DEBUG 10")
    logger.info("test INFO 20")
    logger.warning("test WARNING 30")
    logger.error("test ERROR 40")
    logger.critical("test CRITICAL 50")

    if logger.isEnabledFor(logging.DEBUG):
        logger.warning("debug enabled!")
    if logger.isEnabledFor(LogLevel.DEBUG):
        logger.info("detail enabled")


if __name__ == "__main__":
    _test()
